from flask_login import current_user
from ..models import Question, Answer, Trained, Keyword
from ..stateEnum import QuestionEnum, AnswerEnum
from datetime import datetime
from typing import List
from .. import db

# Get from database functions for database -------------------------------------------------------------------------------------

#Question
def get_question(id: int) -> Question:
    """Return the question with the given id"""
    return Question.query.get(id)

def get_all_questions() -> List[Question]:
    """Returns all questions"""
    return Question.query.all()

#def getQuestionsByTime(time:str)->List[Question]:

    #return list(Question.query.filter_by(timestamp=time))

def get_questions_by_state(state: str) -> List[Question]: 
    """ Function that gets all questions with a given state from the database.
     
    :param state: The string representing the QuestionEnum name or the key word "all". """

    if state == "all":
        return get_all_questions()

    if any(x for x in QuestionEnum if x.name == state):
        return list(Question.query.filter_by(state=QuestionEnum[state].value))
    else:
        print('Error in getQuestionsByState: Invalid input: state')
        return []

def get_questions_by_true_answer(answer: Answer) -> List[Question]:
    """Returns all questions that have the given answer as their true_answer"""
    return list(Question.query.filter_by(true_answer_id=answer.id))

def get_questions_by_bot_answer(answer: Answer) -> List[Question]:
    """Returns all questions that have the given answer as their bot_answer"""
    return list(Question.query.filter_by(bot_answer_id=answer.id))

def get_questions_by_answer(answer: Answer) -> List[Question]:
    """Returns all questions that have the given answer as their bot_answer or true_answer"""
    return list(Question.query.filter((Question.bot_answer_id == answer.id) | (Question.true_answer_id == answer.id)).all())


def get_valid_questions_by_time_stat(start: datetime, end: datetime) -> List[Question]:
    """ Function that gets all VALID and TRAINED_ON questions between the given start and end times from the database.

    :param start: The start datetime object.
    :param end: The end datetime object.
    """
    return Question.query.filter(Question.timestamp >= start, Question.timestamp <= end, Question.state!=QuestionEnum.NOT_VALIDATED.value).all()

def get_valid_question_count_by_true_answer(start_date, end_date) -> dict[str, int]: 
    """ Function that returns a list of dictionaries on the form:
        description - the name of the answer
        count - number of question whit the given answer as true_answer in the given time interval"""
    question_count_by_answer = db.session.query(Question.true_answer_id, Answer.description, db.func.count(Question.id)) \
        .join(Answer, Question.true_answer_id == Answer.id) \
        .filter(Question.timestamp >= start_date) \
        .filter(Question.timestamp <= end_date) \
        .filter(Question.state != QuestionEnum.NOT_VALIDATED.value) \
        .group_by(Question.true_answer_id, Answer.description) \
        .order_by(db.func.count(Question.id).desc()) \
        .all()

    # Create a list of dictionaries containing the answer description and count
    result = [{'description': row[1], 'count': row[2]} for row in question_count_by_answer]

    return result

#def count_question_by_answer(questions: List[Question], answers: List[Answer]) -> List[int]:
    
    #return [len(list(filter(lambda q: q.true_answer_id == answer.id, questions))) for answer in answers] 

#def get_questions_from_last_month() -> List[Question]:
    """# Function for getting all questions from the last month."""
#    end_date = datetime.now()
#    start_date = end_date - timedelta(days=30)
#    stat = get_valid_questions_by_time_stat(start_date, end_date)
#    return len(stat)

#Answer
def get_answer(id: int) -> Answer:
    """Return the answer with the given id"""
    return Answer.query.get(id)

def get_all_answers() -> List[Answer]:
    """Return all questions"""
    return Answer.query.all()

def get_active_answers() -> List[Answer]:
    """Return all active questions, both standard and season"""
    active_standard = list(Answer.query.filter_by(state=AnswerEnum.STANDARD.value))
    active_season = list(Answer.query.filter_by(state=AnswerEnum.SEASON.value))
    answers = active_standard + active_season
    if answers:
        return answers
    else:
        print('Error in getActiveAnswers: No active answers exsist.')
        return []

def get_inactive_answers() -> List[Answer]:
    """Return all inactive questions, both standard and season"""
    inactive_standard = list(Answer.query.filter_by(state=AnswerEnum.INACTIVE_STANDARD.value))
    inactive_season = list(Answer.query.filter_by(state=AnswerEnum.INACTIVE_SEASON.value))
    answers = inactive_standard + inactive_season
    if answers:
        return answers
    else:
        print('Error in getActiveAnswers: No active answers exsist.')
        return []

def get_standard_answers() -> List[Answer]:
    """Return all standard questions, both active and inactive"""
    active_standard = list(Answer.query.filter_by(state=AnswerEnum.STANDARD.value))
    inactive_standard = list(Answer.query.filter_by(state=AnswerEnum.INACTIVE_STANDARD.value))
    answers = active_standard + inactive_standard
    if answers:
        return answers
    else:
        print('Error in getStandardAnswers: No standard answers exsist.')
        return []

def get_season_answers() -> List[Answer]:
    """Return all season questions, both active and inactive"""
    active_season = list(Answer.query.filter_by(state=AnswerEnum.SEASON.value))
    inactive_season = list(Answer.query.filter_by(state=AnswerEnum.INACTIVE_SEASON.value))
    answers = active_season + inactive_season
    if answers:
        return answers
    else:
        print('Error in getSeasonAnswers: No season answers exsist.')
        return []
 
def get_active_season_answers() -> List[Answer]:
    """Reurn all active season answers"""
    answers = list(Answer.query.filter_by(state=AnswerEnum.SEASON.value))
    if answers:
        return answers
    else:
        print('Error in getActiveSeasonAnswers: No active season answers exsist.')
        return []
    
def get_answer_by_label(label: str) -> Answer:
    """Return the first answer with the given label"""
    answer = Answer.query.filter_by(lable=label).first()
    if answer:
        return answer
    else:
        print("Error in getAnswerByLabel: Answer label '"+label+"' does not exsist.")

def get_answer_by_state(state: str) -> List[Answer]:
    """ Function that gets all answers with a given state from the database.
     
    :param state: The string representing the AnswerEnum name or the key word "all". """
    if state == "all":
        return get_all_answers()

    if any(x for x in AnswerEnum if x.name == state):
        return list(Answer.query.filter_by(state=AnswerEnum[state].value))
    else:
        print('Error in getQuestionsByState: Invalid input: state')
        return []
    

#Keyword
def get_keywords_by_answer(answer: Answer) -> List[Keyword]:
    """Return all keyword to the given anser in a list"""
    keyword = Keyword.query.filter_by(answer_id=answer.id)
    if keyword:
        return keyword
    else:
        print('Error in get_keywords_by_answer: No keyword found.')

#Trained
def get_all_trained() -> List[Trained]:
    """Return a list of all trained element"""
    return Trained.query.all()

def get_last_trained() -> Trained:
    """Return the the trained element with the last timestamp"""
    trained_entries = get_all_trained()
    if trained_entries:
        last_trained = trained_entries[0]
        for entry in trained_entries:
            if entry.timestamp > last_trained.timestamp:
                last_trained = entry
        return last_trained
    return Trained(timestamp=datetime(year=1,month=1,day=1),datapoints=0)

# Update database ----------------------------------------------------------------------------------------------------------------

def save_question(question: Question):
    """Save the given question in the database"""
    q = Question.query.get(question.id)
    if q:
        q = question
        db.session.commit()
    else:
        db.session.add(question)
        db.session.commit()

def delete_question(question: Question):
    """Delete the given question from the database"""
    q = Question.query.get(question.id)
    if q:
        Question.query.filter_by(id = q.id).delete()
        db.session.commit()
    else:
        print('deleted question not found')

def add_trained_entry(num_of_train_data: int):
    """Create and save a trained element with numOfTrainData and curent datetime in the database"""
    user_id = current_user.id
    timestamp = datetime.now()
    new_trained = Trained(timestamp=timestamp,datapoints=num_of_train_data, user_id=user_id)
    db.session.add(new_trained)
    db.session.commit()

def add_keyword(answer: Answer, keyword: str):
    """Create and save a keyword element whit the given answer and keyword in the database"""
    new_keyword = Keyword(word=keyword,answer_id=answer.id)
    db.session.add(new_keyword)
    db.session.commit()

def remove_keyword(keyword: Keyword):
    """Delete the given keyword from the database"""
    k = Keyword.query.get(keyword.id)
    if k:
        Keyword.query.filter_by(id = k.id).delete()
        db.session.commit()
    else:
        print('Error in removeKeyword: Keyword to be removed not found')

def deactivate_answer(answer: Answer):
    """ Changes the state of the answer STANDARD to INACTIVE_STANDARD and SEASON to INACTIVE_SEASON, and changes the state of all changed questions to NOT_VALIDATED. """
    questions = get_questions_by_true_answer(answer)
    for question in questions:
        question.state = QuestionEnum.NOT_VALIDATED.value
        db.session.commit()
    if answer.state == AnswerEnum.STANDARD.value:
        answer.state = AnswerEnum.INACTIVE_STANDARD.value
        db.session.commit()
    elif answer.state == AnswerEnum.SEASON.value:
        answer.state = AnswerEnum.INACTIVE_SEASON.value
        db.session.commit()
    else:
        print('Error in deactivateAnswer: Answer already deactivated')

def activate_answer(answer: Answer):
    """ Changes the state of the answer INACTIVE_STANDARD to STANDARD and INACTIVE_SEASON to SEASON. """
    if answer.state == AnswerEnum.INACTIVE_STANDARD.value:
        answer.state = AnswerEnum.STANDARD.value
        db.session.commit()
    elif answer.state == AnswerEnum.INACTIVE_SEASON.value:
        answer.state = AnswerEnum.SEASON.value
        db.session.commit()
    else:
        print('Error in activateAnswer: Answer already activated')

def delete_answer(answer: Answer) -> bool:
    """ Deletes an answer and all keywords pointing to it. You can not delete an answer where a questions true answer it pointing to it. If this is a problem, considre using the func changeTrueAnswer """
    questions = get_questions_by_answer(answer)
    keywords = get_keywords_by_answer(answer)

    if len(questions)>0:
        print('Error in deleteAnswer: You can not delete an answer with questions pointing to it')
    else:
        for keyword in keywords : remove_keyword(keyword)
        a = Answer.query.get(answer.id)
        if a:
            Answer.query.filter_by(id = a.id).delete()
            db.session.commit()
            return True
        else:
            print('Error in deleteAnswer: Deleted answer not found')
    return False

def change_state_of_question_to(questions: List[Question], state: int):
    """ Changes the state of the questions to the state with the input value """
    for question in questions:
        question.state = state
        db.session.commit()

def change_true_answer(fro: Answer, to: Answer):
    """ Gets all questions where true answer is the same as "fro", and changes them to "to" """
    questions = get_questions_by_true_answer(fro)
    for question in questions:
        question.true_answer_id = to.id