from ..models import Question, Answer
from ..stateEnum import QuestionEnum, AnswerEnum
from .database import *
from datetime import datetime
from typing import List
import re

# Sorting and filtring functions ------------------------------------------------------------------------------------------------
def is_validated(question: Question) -> bool:
    """ Function that checks if a question is validatet or not. """
    if question.state == QuestionEnum.VALIDATED.value:
        return True
    else:
        return False

def is_trained_on(question: Question) -> bool:
    """ Function that checks if a question is trained on or not. """
    if question.state == QuestionEnum.TRAINED_ON.value:
        return True
    else:
        return False

def is_active(answer: Answer) -> bool:
    """Return true if the given answer is active"""
    if answer.state == AnswerEnum.STANDARD.value:
        return True
    elif answer.state == AnswerEnum.SEASON.value:
        return True
    elif answer.state == AnswerEnum.INACTIVE_STANDARD.value:
        return False
    elif answer.state == AnswerEnum.INACTIVE_SEASON.value:
        return False
    else:
        print('Error in isActive: Answer state unknown')
        return False

def sort_questions_by(questions: List[Question], sort_key: str) -> List[Question]:
    """ Function that takes a list and sorts it with the given sort_key.
     
    :param questions: List of questions you want sorted.
    :param sort_key: String 'timestamp' or 'category' based on what you want to sort the list on. """
    if sort_key == 'timestamp_last_first':
        return sorted(questions, key=lambda x: x.timestamp, reverse=True)
    elif sort_key == 'timestamp_first_last':
        return sorted(questions, key=lambda x: x.timestamp)
    elif sort_key == 'bot_answer':
        return sorted(questions, key=lambda x: x.bot_answer_id)
    else:
        print('Error in sortQuestionsBy: Invalid input: sort_key')
        return []
    
def filter_questions_with_timestamp(questions: List[Question], start: datetime = None, end: datetime = None) -> List[Question]:
    """ Function that takes a list of questions, and optionaly a starttime and endtime. Function returns a list of questions between the start and endtime.
     
    :param questions: List of questions.
    :param start: Optional datetime object. To use write 'start = DATETIME_OBJECT' 
    :param end: Optional datetime object. To use write 'end = DATETIME_OBJECT' """
    filteredList = []
    for question in questions:
        time = question.timestamp
        if (not start or time>=start) and (not end or time<=end):
            filteredList.append(question)
    return filteredList

def filter_questions_with_bot_answer(questions: List[Question], answer_label: str) -> List[Question]:
    """ Function that takes a list of question, and returns all the questions where the lable of the bot answer equals the inputed answer_label. """
    if answer_label == "all":
        return questions
    
    filterdQuestions = []
    for question in questions:
        bot_answer = get_answer(question.bot_answer_id)
        if bot_answer.lable == answer_label:
            filterdQuestions.append(question)
    return filterdQuestions

def filter_by_answer_lable(questions: List[Question], filter_lable: str) -> List[Question]:
    """if filter_lable is all then return all questions or return a list of question with true_anser equal to the answer with the given filter_lable"""
    if filter_lable != 'all':
        filterByAnswerId = get_answer_by_label(filter_lable).id
        return [question for question in questions if question.true_answer_id == filterByAnswerId]
    else:
        return questions

#def getNumQuestionsFromLastMonth() -> int:
#    """Returns the number of questions in the last month."""
#    return len(get_questions_from_last_month())

def is_season(question: str) -> int:
    """ Returns the id of the first answer with a keyword the input question contains. If no answers have a matching keyword it returns None """
    answers = get_active_season_answers()
    for answer in answers:
        keywords = get_keywords_by_answer(answer)
        for keyword in keywords:
            if keyword.word.lower() in question.lower():
                return answer.id
    return None


def is_safe(answer: str) -> bool:
    """Return true if the given answer dose not contains unsafe words"""
    windowEvents = ["onafterprint", "onbeforeprint", "onbeforeunload", "onerror", "onhaschange", "onload", "onmessage", "onoffline", "ononline", "onpagehide", "onpageshow", "onpopstate", "onresize", "onstorage", "onunload"]
    formEvents = ["onblur", "onchange", "oncontextmenu", "onfocus", "oninput", "oninvalid", "onreset", "onsearch", "onselect", "onsubmit"]
    keybordEvents = ["onkeydown", "onkeypress", "onkeyup"]
    mouseEvents = ["onclick", "ondblclick", "onmousedown", "onmousemove", "onmouseout", "onmouseover", "onmouseup", "onmousewheel", "onwheel"]
    dragEvents = ["ondrag", "ondragend", "ondragenter", "ondragleave", "ondragover", "ondragstart", "ondrop", "onscroll"]
    clipboardEvents = ["oncopy", "oncut", "onpaste"]
    mediaEvents = ["onabort", "oncanplay", "oncanplaythrough", "oncuechange", "ondurationchange", "onemptied", "onended", "onloadeddata", "onloadedmetadata", "onloadstart", "onpause", "onplay", "onplaying", "onprogress", "onratechange", "onseeked", "onseeking", "onstalled", "onsuspend", "ontimeupdate", "onvolumechange", "onwaiting"]
    otherEvents = ["ontoggle"]

    events = windowEvents + formEvents + keybordEvents + mouseEvents + dragEvents + clipboardEvents + mediaEvents + otherEvents
    digits = ["[","]","{","}"]
    words = ["script", "Script"]

    unsafe = events + digits + words

    # Checkes if the remaining string contains unsafe words
    for word in unsafe:
        if word in answer:
            return False
    return True

def to_lable(description: str) -> str:
    """Filter out not allowed characters"""
    lable = re.sub('[^a-zA-Z0-9\n]', '', description)
    return lable

# Used to correct labels from datadumps
def correct_label(label: str) -> str:
        """Fix typos in label"""
        answer_dict = {
        #Feil skrevet kattegorier
        'buy_tickets': 'buy_tickets',
        'program_content': 'program_content',
        'gavekort': 'gavekort',
        'festspillgavekort': 'gavekort',
        'get_dress_code': 'get_dress_code',
        'find_festival_dates': 'find_festival_dates',
        'find_prices': 'find_prices',
        'get_refund': 'get_refund',
        'unknown ': 'unknown',
        'true': 'true',
        'unknown': 'unknown',
        'get_bicycle': 'get_bicycle',
        'print_tickets': 'print_tickets',
        'volunteer': 'volunteer',
        'find_festival dates': 'find_festival_dates',
        'ticketing_request': 'ticketing_request',
        'program_content ': 'program_content',
        'FiB Fordel-info ': 'find_rebates',
        'unkown': 'unknown',
        # np.nan: np.nan,
        'find_rebates': 'find_rebates',
        'get_rebates': 'find_rebates',
        'festspillkort': 'festspillkort',
        'Unknown': 'unknown',
        'true ': 'true',
        #Slå saman kattegorier
        'ticketing_request': 'buy_tickets',
        'find_rebates': 'find_prices',
        }

        #Fikse opp i kategoriane
        for key, value in answer_dict.items():
            label = label.replace(key, value)
        return label

def toggle_answer_state(answer: Answer):
    """Chache the state of answer to the opposite state"""
    if is_active(answer):
        deactivate_answer(answer)
    else:
        activate_answer(answer)